#include<iostream>
#include<fstream>
#include<cstring>
#include<cmath>
#include<sstream>
#include<ctime>

#include<TGraph.h>
#include<TGraphErrors.h>
#include<TGraph2D.h>
#include<TGraph2DErrors.h>
#include<TCanvas.h>
#include<TFile.h>
#include<TH1F.h>
#include<TF2.h>
#include<TTree.h>
#include<TTimer.h>
#include<TLegend.h>
#include<TStyle.h>	

using namespace std;

// time converter (time string to seconds since 1970)

long long time_to_sec(string date, string time, string meridiem) {
   string stime = date + " " + time + " " + meridiem;

   struct tm T;
   strptime (stime.c_str(), "%m/%d/%Y %I:%M:%S %p", &T);

   cout << "Event time: " << stime.c_str() << endl;
   
   long long epoch = (long long) mktime(&T); 
//   cout << "Event time (epoch) " <<  epoch << endl;
   return epoch;  
}


// hexadecimal to binary converter

string hextobin(char bin) {
   string bit;
   
   switch(bin) {
      case '0': bit = "0000"; break;
      case '1': bit = "0001"; break;
      case '2': bit = "0010"; break;
      case '3': bit = "0011"; break;
      case '4': bit = "0100"; break;
      case '5': bit = "0101"; break;
      case '6': bit = "0110"; break;
      case '7': bit = "0111"; break;
      case '8': bit = "1000"; break;
      case '9': bit = "1001"; break;
      case 'A': bit = "1010"; break;
      case 'B': bit = "1011"; break;
      case 'C': bit = "1100"; break;
      case 'D': bit = "1101"; break;
      case 'E': bit = "1110"; break;
      case 'F': bit = "1111"; break;
      case 'a': bit = "1010"; break;
      case 'b': bit = "1011"; break;
      case 'c': bit = "1100"; break;
      case 'd': bit = "1101"; break;
      case 'e': bit = "1110"; break;
      case 'f': bit = "1111"; break;
    }
    
    return bit;          
} 

// Least square linear regression for tracks on planes xz and yz

void LS (vector<double> x,  vector<double> y, double &m, double &q) { // linear least squares coefficients 

   double mx = 0; 
   double my = 0; 

   int size_x = x.size();
 
   for (int k = 0; k < size_x; k++) {
      mx += x[k];
      my += y[k];      
   }   
   
   mx = mx/size_x;
   my = my/size_x;

   double sx = 0; 
   double cxy = 0;
  
   for (int k = 0; k < size_x; k++) sx += pow(x[k]  - mx, 2);
   for (int k = 0; k < size_x; k++) cxy += (x[k]  - mx)*(y[k]  - my);

   cxy = cxy/size_x;
   sx = sx/size_x;
   
   m = cxy/sx;
   q = my - m*mx;
}

// *** DECODER ***

void decoder(string file, string fileout = "out.root", int nmax = 1e9) {
   
   ifstream fin(file.c_str());
   string n, x, y, date, time, meridiem;
   
   TFile *fout = new TFile(fileout.c_str(), "RECREATE");
   
   // reconstructed event
   TTree *events = new TTree(); events->SetName("events");
   
   int cut, cut_x, cut_y, event;    
   double theta, phi; // standard polar angles
   double theta_x, theta_y; // Gustavino's agles
   int pos_x[24], pos_y[24];
   double chi2, chi2_x, chi2_y;
   long long time_sec; 
   double X0, Y0, Z0; 


   events->Branch("event", &event, "event/I"); 
   events->Branch("time_sec", &time_sec, "time_sec/l");                            
   events->Branch("pos_x", &pos_x, "pos_x[24]/I");
   events->Branch("pos_y", &pos_y, "pos_y[24]/I");
   events->Branch("cut", &cut, "cut/I");       
   events->Branch("cut_x", &cut_x, "cut_x/I");       
   events->Branch("cut_y", &cut_y, "cut_y/I");       
   events->Branch("theta", &theta, "theta/D");       
   events->Branch("phi", &phi, "phi/D");   
   events->Branch("theta_x", &theta_x, "theta_x/D");
   events->Branch("theta_y", &theta_y, "theta_y/D");     
   events->Branch("chi2", &chi2, "chi2/D");     
   events->Branch("chi2_x", &chi2_x, "chi2_x/D");     
   events->Branch("chi2_y", &chi2_y, "chi2_y/D");     
   events->Branch("X0", &X0, "X0/D");  
   events->Branch("Y0", &Y0, "Y0/D");  
   events->Branch("Z0", &Z0, "Z0/D");  

   TTree *meta_event = new TTree(); meta_event->SetName("meta_event");
   
   double livetime;
   meta_event->Branch("livetime", &livetime, "livetime/D");     
      
   // detectir size          
   double scale = 1;     
   double h = scale*7./4;
   double floor[4] = {3*h, 2*h, 1*h, 0*h}; 
 
   // control plots 
   TH1F* ftheta = new TH1F("ftheta", "#theta distribution",30,  0, 60);
   TH1F* fphi = new TH1F("fphi", "#phi distribution", 30, 0, 360);
   ftheta->SetLineWidth(2);
   fphi->SetLineWidth(2);   
 
   TH1F* ftheta_x = new TH1F("ftheta_x", "#theta_{x} distribution",30,  -60, 60);
   TH1F* ftheta_y = new TH1F("ftheta_y", "#theta_{y} distribution",30,  -60, 60);
   ftheta_x->SetLineWidth(2);
   ftheta_y->SetLineWidth(2);   
 
       
   TCanvas* ctf = new TCanvas("distributions", "distributions", 1);
   ctf->Divide(2,2);  

   // loop    
   event = 0;  
   long long t1 = 0, t2 = 0;

   bool first = true;
   while(fin && event < nmax) { // FIXME for efficiency only
      cout << "-->Event no. " << event  << endl;
       
      // input raw rile data  
      vector<double> vx, vy, vzx, vzy;     
      fin >> n >>  x >> y >> date >> time >> meridiem;
//      fin >>  x >> y >> date >> time >> am;
  
      if(!fin) break;
      
      t2 = time_to_sec(date, time, meridiem);
      time_sec = t2;
      
      if (first) {
         t1 = t2;
         first = false;     
      } 
      cout << "Epoch: " <<  t1 << " - " << t2 << endl;


//      cout << "x: " << x <<  " - y: "  << y << endl;
//      cout << "decoded : " << endl;
      
      cut_x = 0;
      cut_y = 0;
      
      int good[11] = {1, 2, 4, 8, 16, 32, 3, 6, 12, 24, 48};  // cluster size = 1 or = 2, accepted
   
      // layers
      for (int i = 0; i < 4; i++) { // exclude 3rd floor from bottom
      
         string xx = hextobin(x[0 + i*2]) + hextobin(x[1 + i*2]);
         string yy  = hextobin(y[0 + i*2]) + hextobin(y[1 + i*2]);
         
         string sx = xx.substr(2,6); 
         
         if (i == 1) sx[4] = '0'; // turn off broken pixel
                   
         string sy = yy.substr(2,6);
         
         cout << sx << " - " << sy << endl;
         
         int sum_x = 0;
         int sum_y = 0;
           
         for (int j = 0; j < 6; j++) {
     
            pos_x[j + i*6] = sx[j] - 48;
            pos_y[j + i*6] = sy[j] - 48;

            if(pos_x[j + i*6]) { 
              vzx.push_back(floor[i]);
              vx.push_back(j + 0.5);
            }
               
            if(pos_y[j + i*6]) { 
              vzy.push_back(floor[i]);
              vy.push_back(j + 0.5);
            }

            sum_x += pos_x[j + i*6]*pow(2, j);
            sum_y += pos_y[j + i*6]*pow(2, j); 
         }   

         for (int k = 0; k < 11; k++) {
            if (sum_x == good[k]) cut_x += 1;
            if (sum_y == good[k]) cut_y += 1;
         }

         //if (i == 1 && sum_x == 0) cut_x += 1; // make exception for 2nd layer from top. Removed by Gustavino's razor


      if (cut_x == 4 && cut_y == 4) cut = 1;
      else cut = 0;
 
      } 
      event++;  
      // track reconstrution       
 
      double mx, my, qx, qy;

      chi2_x = 0; 
      chi2_y = 0;  
     
      if (vx.size() > 1  && vy.size() > 1) {
         LS(vzx, vx, mx, qx);
         LS(vzy, vy, my, qy);

         for (size_t j = 0; j < vx.size(); j++) chi2_x += pow(vzx[j] - mx*vx[j] - qx, 2)/12/2; 
         for (size_t j = 0; j < vy.size(); j++) chi2_y += pow(vzy[j] - my*vy[j] - qy, 2)/12/2; 
         chi2 = (chi2_x + chi2_y)/2;
      
         double vz = 1; 
         theta = acos(vz/sqrt(mx*mx + my*my + vz*vz))*180/M_PI;
         phi = atan2(my,mx)*180/M_PI + 180;
         theta_x = atan(mx/vz)*180/M_PI;
         theta_y = atan(my/vz)*180/M_PI;
         
         Z0 = 3*h;
         X0 = Z0*mx;
         Y0 = Z0*my;
      }   
      else {
         theta = 0; 
         phi = 0;
         theta_x = 0; 
         theta_y = 0; 
         cut = 0;
      }
           
//      cout << "theta = " << theta << " - phi = " << phi << endl;
//      cout << "theta_x = " << theta_x << " - theta_y = " << theta_y << endl;
//      cout << "Event quality cut (4 == good): " << cut << endl;
//      cout << endl; 

      events->Fill();
 
      if(cut == 1) {
         ftheta->Fill(theta);
         fphi->Fill(phi);
         ftheta_x->Fill(theta_x);
         ftheta_y->Fill(theta_y);         
      }         
   }

   ctf->cd(1);    
   ftheta->Draw();  
   
   ctf->cd(2);
   fphi->Draw();

   ctf->cd(3);    
   ftheta_x->Draw();  
   
   ctf->cd(4);
   ftheta_y->Draw();
   
   ctf->Write();
 
   events->Write();
   
   //livetime calculation
   
   livetime = t2 - t1;
   cout << "Live time: " <<  livetime << " sec" << endl;   
   
   meta_event->Fill();
   meta_event->Write();
      
   fout->Close();
   
}   


// viewer

void view(string rootfile) { // 3D event display

   gStyle->SetTitleFontSize(0.06);
   gStyle->SetTitleSize(0.06, "xy");
   gStyle->SetTitleOffset(0.8, "y");   
   gStyle->SetTitleOffset(0.8, "x");   
   gStyle->SetLabelSize(0.06, "xy"); 

   TFile *fin = new TFile(rootfile.c_str(), "READ");
   TTree *events = (TTree*)fin->Get("events");

   int cut, cut_x, cut_y, event;
   double theta, phi; // standard polar angles
   double theta_x, theta_y; // Gustavino's agles
   int pos_x[24], pos_y[24];
   
   events->SetBranchAddress("event", &event);             
   events->SetBranchAddress("pos_x", &pos_x[0]);
   events->SetBranchAddress("pos_y", &pos_y[0]);
   events->SetBranchAddress("cut", &cut);       
   events->SetBranchAddress("theta", &theta);       
   events->SetBranchAddress("phi", &phi);   
   events->SetBranchAddress("theta_x", &theta_x);
   events->SetBranchAddress("theta_y", &theta_y);  
   events->SetBranchAddress("cut_x", &cut_x);  
   events->SetBranchAddress("cut_y", &cut_y);  
   
   TTimer *timer = new TTimer("gSystem->ProcessEvents();", 10000, kFALSE); 

   double scale = 1;
   double h = 7./4*scale;
   double floor[4] = {3*h, 2*h, 1*h, 0*h};   
   double vz = 1;
   double z1 = 3*h;
      
   TCanvas* c = new TCanvas("c", "c", 1000, 1000);
   c->Divide(2,2);
   
   int k = 0;
   while(1) {
      events->GetEntry(k);
         
      vector<double> vx, vy, vzx, vzy, evx, evy, evz;  

      for (int i = 0; i < 4; i++) {
         for (int j = 0; j < 6; j++) {
 
            if(pos_x[j + i*6]) { 
              vzx.push_back(floor[i]);
              vx.push_back(j + 0.5);
              evx.push_back(1./sqrt(12));
              evz.push_back(0);
            }
               
            if(pos_y[j + i*6]) { 
              vzy.push_back(floor[i]);
              vy.push_back(j + 0.5);
              evy.push_back(1./sqrt(12));
              evz.push_back(0);              
            }
         }  
      } 

      // reconstructed 3D       

      double mx, my, qx, qy;
      
      if (vx.size() > 1  && vy.size() > 1) {
         LS(vzx, vx, mx, qx);
         LS(vzy, vy, my, qy);
      }   
      else {
         mx = 0;
         my = 0;
         qx = 0;
         qy = 0;
      }   
 
      TGraph2D *gp = new TGraph2D();
      gp->SetName("GP");
      gp->SetPoint(0, 0, 0, 0);
      gp->SetPoint(1, 6, 6, z1);
      gp->SetTitle(Form("Event #%d; x view; y view; z", k));   

      TGraph *gpx = new TGraph();
      gpx->SetName("GP");
      gpx->SetPoint(0, 0, 0);
      gpx->SetPoint(1, 6, z1);
      gpx->SetTitle(Form("Event #%d; x view; z", k));   

      TGraph *gpy = new TGraph();
      gpy->SetName("GP");
      gpy->SetPoint(0, 0, 0);
      gpy->SetPoint(1, 6, z1);
      gpy->SetTitle(Form("Event #%d; y view; z", k));   

   
      TGraph2DErrors *g = new TGraph2DErrors(vx.size(), &vx[0], &vy[0], &vzx[0], &evx[0], &evy[0], &evz[0]); 
      g->SetName("G");
      g->SetMarkerStyle(20);
      g->SetMarkerSize(1.5);
      g->SetLineWidth(3);

      TGraphErrors *gx = new TGraphErrors(vx.size(), &vx[0], &vzx[0], &evx[0], &evz[0]); 
      gx->SetName("GX");
      gx->SetMarkerStyle(20);
      gx->SetLineWidth(3);
      gx->SetMarkerSize(1.5);

      TGraphErrors *gy = new TGraphErrors(vy.size(), &vy[0], &vzy[0], &evy[0], &evz[0]); 
      gy->SetName("GY");
      gy->SetMarkerStyle(20);
      gy->SetMarkerSize(1.5);
      gy->SetLineWidth(3);


      TGraph2D *gt = new TGraph2D();
      gt->SetName("GT");
      gt->SetLineColor(kRed);
      gt->SetLineWidth(4);     
      gt->SetPoint(0, qx, qy, 0);
      gt->SetPoint(1, qx + mx*z1, qy + my*z1, z1);
 
      TGraph *gtx = new TGraph();
      gtx->SetName("GTX");
      gtx->SetLineColor(kBlue);
      gtx->SetLineWidth(4);     
      gtx->SetPoint(0, qx, 0);
      gtx->SetPoint(1, qx + mx*z1, z1);

      TGraph *gty = new TGraph();
      gty->SetName("GTY");
      gty->SetLineColor(kGreen+1);
      gty->SetLineWidth(4);     
      gty->SetPoint(0, qy, 0);
      gty->SetPoint(1, qy + my*z1, z1);

      c->cd(3);
      
      gp->GetXaxis()->SetTitleOffset(1.2);
      gp->GetYaxis()->SetTitleOffset(1.2);
      
      gp->Draw("P"); 
      g->Draw("Perrsame");
      gt->Draw("LINEsame");

      c->cd(1);
      gpx->GetYaxis()->SetRangeUser(-0.5, 6);
      gpx->Draw("AP"); 
      gx->Draw("Psame");
      gtx->Draw("Lsame");

      c->cd(2);
      gpy->GetYaxis()->SetRangeUser(-0.5, 6);
      gpy->Draw("AP"); 
      gy->Draw("Psame");
      gty->Draw("Lsame");

      c->cd(4);

      TLegend *leg = new TLegend(0.1, 0.1, 0.9, 0.9);
      leg->AddEntry(g,"points","p");
      leg->AddEntry(gt,"track 3D","l");
      leg->AddEntry(gtx,"track X","l");
      leg->AddEntry(gty,"track Y","l");
      leg->AddEntry("", Form("#theta = %2.2f",theta), "");
      leg->AddEntry("", Form("#phi = %2.2f",phi), "");
      leg->AddEntry("", Form("#theta_{x} = %2.2f",theta_x), "");
      leg->AddEntry("", Form("#theta_{y} = %2.2f",theta_y), "");
      leg->AddEntry("", Form("cut = %d",cut), "");
      leg->AddEntry("", Form("cut_x = %d",cut_x), "");
      leg->AddEntry("", Form("cut_y = %d",cut_y), "");

      leg->Draw("same");

      c->Modified();
      c->Update();
      k++;
      
      timer->TurnOn(); // for intective modification of the canvas
      timer->Reset();   

      string go;
      cout << "RETURN to continue: ";
      getline(cin, go);  
      if (go == "") {
         delete leg; 
         delete gt;
         delete g;
         delete gpx;
         delete gtx;
         delete gx;
         delete gpy;
         delete gty;
         delete gy;
         delete gp;

         continue;
      }
      else return;

      timer->TurnOff();
   }  
}

// *** main() ***

int main(int argc, char *argv[]) {
   
   if (argc <= 1) {
      cout << "usage: ./decoder raw_file fileout (optional) nmax (optional)" << endl;
      exit(1); 
   } 


   if (argc == 2) {
      string file  = (string) argv[1];    
      decoder(file);
   }

   if (argc == 3) {
      string file  = (string) argv[1];    
      string fileout  = (string) argv[2];    
      decoder(file, fileout);
   }   

   if (argc == 4) {
      string file  = (string) argv[1];    
      string fileout  = (string) argv[2];    
      int nmax = atoi(argv[3]); 
      decoder(file, fileout, nmax);
   }   


   return 0; 
}


